#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "coprthr.h"

#define SIZE 1048576
#define SIZE16 "1048576 / 16"

char src[] = \
	"#include <coprthr.h>\n" \
	"void my_kern( float* a, float* b, float* c) {\n" \
	"	int idx = coprthr_get_thread_index() * " SIZE16 ";\n" \
	"	int to = idx+" SIZE16 ";\n" \
	"	for(int i=idx; i<to; i++) {\n" \
	"		c[i] = a[i]*b[i];\n" \
	"	}\n" \
	"}\n";

double walltime()
{
	static struct timeval t;
	gettimeofday(&t, NULL);
	return (t.tv_sec + 1e-6 * t.tv_usec);
}

int main()
{
	int i;

	int dd = coprthr_dopen(COPRTHR_DEVICE_E32, COPRTHR_O_STREAM);

	if (dd<0) {
		printf("device open failed\n");
		exit(-1);
	}


	coprthr_program_t prg = coprthr_dcompile(dd, src, sizeof(src), "", 0);
	coprthr_kernel_t krn = coprthr_getsym(prg, "my_kern");

	float* a = (float*) malloc(SIZE*sizeof(float));
	float* b = (float*) malloc(SIZE*sizeof(float));
	float* c = (float*) malloc(SIZE*sizeof(float));

	for(i=0; i<SIZE; i++) {
		a[i] = 1.0f * i;
		b[i] = 2.0f * i;
		c[i] = 0.0f;
	}

	coprthr_mem_t mema = coprthr_dmalloc(dd, SIZE*sizeof(float), 0);
	coprthr_mem_t memb = coprthr_dmalloc(dd, SIZE*sizeof(float), 0);
	coprthr_mem_t memc = coprthr_dmalloc(dd, SIZE*sizeof(float), 0);

	coprthr_dwrite(dd, mema, 0, a, SIZE*sizeof(float), COPRTHR_E_NOWAIT);
	coprthr_dwrite(dd, memb, 0, b, SIZE*sizeof(float), COPRTHR_E_NOWAIT);
	coprthr_dwrite(dd, memc, 0, c, SIZE*sizeof(float), COPRTHR_E_NOWAIT);

	unsigned int nargs = 3;
	void* args[] = { &mema, &memb, &memc };
	unsigned int nthr = 16;

	double runtime;
	double start = walltime();

	coprthr_dexec(dd, krn, nargs, args, nthr, 0, COPRTHR_E_NOWAIT);

	coprthr_dcopy(dd, memc, 0, memb, 0, SIZE*sizeof(float), COPRTHR_E_NOWAIT);

	coprthr_dread(dd, memc, 0, c, SIZE*sizeof(float), COPRTHR_E_NOWAIT);

	coprthr_dwait(dd);

	runtime = walltime() - start;
	printf("Runtime: %8.3f ms\n", runtime*1e3);

	//for(i=0; i<SIZE; i++)
	//	printf("%f * %f = %f\n", a[i], b[i], c[i]);

	coprthr_dfree(dd, mema);
	coprthr_dfree(dd, memb);
	coprthr_dfree(dd, memc);

	free(a);
	free(b);
	free(c);

	coprthr_dclose(dd);
}
