#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <stdint.h>
#include <time.h>

#define FRAMES 100
#define XRES 180
#define YRES 180

unsigned int xres, yres;
unsigned int pixels[2][YRES][XRES];

static inline void nano_wait(uint32_t sec, uint32_t nsec)
{
	struct timespec ts;
	ts.tv_sec = sec;
	ts.tv_nsec = nsec;
	nanosleep(&ts, NULL);
}

unsigned int fetch_node(unsigned int page, int x, int y, int x_diff, int y_diff) {
	int x_new = x + x_diff;
	if (x_new < 0 || x_new >= xres) {
		return 0;
	}
	int y_new = y + y_diff;
	if (y_new < 0 || y_new >= yres) {
		return 0;
	}
	return pixels[page][y_new][x_new];
}

int main(int argc, char *argv[])
{
	struct timespec time;
	double time0, time1;
	xres = XRES;
	yres = YRES;

	unsigned int row = 0;
	unsigned int col = 0;

	FILE *file = fopen("grid.txt", "r");
	if (file == NULL) {
		fprintf(stderr, "Could not open grid.txt\n");
		exit(1);
	}

	int x, y, c;
	int eol = 0;
	for (y = 0; y < yres; y++) {
		for (x = 0; x < xres; x++) {
			if (!eol) {
				c = fgetc(file);
				if (c == '\n') {
					eol = 1;
				}
			}
			if (c == '1') {
				pixels[0][y][x] = 1;
			} else {
				pixels[0][y][x] = 0;
			}
			pixels[1][y][x] = 0;
		}
		eol = 0;
	}

	clock_gettime(CLOCK_REALTIME, &time);
	time0 = time.tv_sec + time.tv_nsec * 1.0e-9;

	unsigned int frame = 1;
	unsigned int page_last = 0;
	unsigned int page_cur = 1;
	while (1) {
		if (frame > FRAMES) {
			break;
		}
		frame++;

		for (y = 0; y < yres; y++) {
			for (x = 0; x < xres; x++) {
				int is_live = fetch_node(page_last, x, y, 0, 0);
				int nr_live = 0;
				nr_live += fetch_node(page_last, x, y, -1, -1);
				nr_live += fetch_node(page_last, x, y, -1, 0);
				nr_live += fetch_node(page_last, x, y, -1, 1);
				nr_live += fetch_node(page_last, x, y, 0, -1);
				nr_live += fetch_node(page_last, x, y, 0, 1);
				nr_live += fetch_node(page_last, x, y, 1, -1);
				nr_live += fetch_node(page_last, x, y, 1, 0);
				nr_live += fetch_node(page_last, x, y, 1, 1);
				if (is_live) {
					if (nr_live < 2 || nr_live > 3) {
						pixels[page_cur][y][x] = 0;
					} else {
						pixels[page_cur][y][x] = 1;
					}
				} else if (nr_live == 3) {
					pixels[page_cur][y][x] = 1;
				} else {
					pixels[page_cur][y][x] = is_live;
				}
			}
		}

		page_last = page_cur;
		page_cur = page_cur ^ 1;
	}

	clock_gettime(CLOCK_REALTIME, &time);
	time1 = time.tv_sec + time.tv_nsec * 1.0e-9;
	printf("rows: %d, cols: %d\n", yres, xres);
	printf("frames: %d\n", FRAMES);
	printf("time: %f sec\n", time1 - time0);

	for (y = 0; y < yres; y++) {
		for (x = 0; x < xres; x++) {
			printf("%d ", pixels[page_last][y][x]);
		}
		printf("\n");
	}

	return 0;
}
