#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <stdint.h>
#include <time.h>

#define FRAMES 100
#define XRES 180
#define YRES 180

unsigned int xres, yres;

static inline void nano_wait(uint32_t sec, uint32_t nsec)
{
	struct timespec ts;
	ts.tv_sec = sec;
	ts.tv_nsec = nsec;
	nanosleep(&ts, NULL);
}

unsigned int fetch_node(char *src, int x, int y, int x_diff, int y_diff) {
	int x_new = x + x_diff;
	if (x_new < 0 || x_new >= xres) {
		return 0;
	}
	int y_new = y + y_diff;
	if (y_new < 0 || y_new >= yres) {
		return 0;
	}
	char *src2 = src + (y_new * yres + x_new) * sizeof(unsigned int);
	return *src2;
}

int main(int argc, char *argv[])
{
	struct timespec time;
	double time0, time1;
	xres = XRES;
	yres = YRES;

	unsigned int row = 0;
	unsigned int col = 0;

	char *src[2] = {
		malloc(xres * yres * sizeof(unsigned int)),
		malloc(xres * yres * sizeof(unsigned int))
	};

	FILE *file = fopen("grid.txt", "r");
	if (file == NULL) {
		fprintf(stderr, "Could not open grid.txt\n");
		exit(1);
	}

	int x, y, c;
	int eol = 0;
	unsigned int *pixel_last = (unsigned int *)src[0];
	unsigned int *pixel_cur = (unsigned int *)src[1];
	for (y = 0; y < yres; y++) {
		for (x = 0; x < xres; x++) {
			if (!eol) {
				c = fgetc(file);
				if (c == '\n') {
					eol = 1;
				}
			}
			if (c == '1') {
				*pixel_last = 1;
			} else {
				*pixel_last = 0;
			}
			*pixel_cur = 0;
			pixel_last++;
			pixel_cur++;
		}
		eol = 0;
	}

	clock_gettime(CLOCK_REALTIME, &time);
	time0 = time.tv_sec + time.tv_nsec * 1.0e-9;

	unsigned int frame = 1;
	unsigned int page = 1;
	while (1) {
		pixel_cur = (unsigned int *)src[page];
		page = page ^ 1;

		if (frame > FRAMES) {
			break;
		}
		frame++;

		for (y = 0; y < yres; y++) {
			for (x = 0; x < xres; x++) {
				int is_live = fetch_node(src[page], x, y, 0, 0);
				int nr_live = 0;
				nr_live += fetch_node(src[page], x, y, -1, -1);
				nr_live += fetch_node(src[page], x, y, -1, 0);
				nr_live += fetch_node(src[page], x, y, -1, 1);
				nr_live += fetch_node(src[page], x, y, 0, -1);
				nr_live += fetch_node(src[page], x, y, 0, 1);
				nr_live += fetch_node(src[page], x, y, 1, -1);
				nr_live += fetch_node(src[page], x, y, 1, 0);
				nr_live += fetch_node(src[page], x, y, 1, 1);
				if (is_live) {
					if (nr_live < 2 || nr_live > 3) {
						*pixel_cur = 0;
					} else {
						*pixel_cur = 1;
					}
				} else if (nr_live == 3) {
					*pixel_cur = 1;
				} else {
					*pixel_cur = is_live;
				}

				pixel_cur++;
			}
		}
	}

	clock_gettime(CLOCK_REALTIME, &time);
	time1 = time.tv_sec + time.tv_nsec * 1.0e-9;
	printf("rows: %d, cols: %d\n", yres, xres);
	printf("frames: %d\n", FRAMES);
	printf("time: %f sec\n", time1 - time0);

	pixel_cur = (unsigned int *)src[page];
	for (y = 0; y < yres; y++) {
		for (x = 0; x < xres; x++) {
			printf("%d ", *pixel_cur);
			pixel_cur++;
		}
		printf("\n");
	}

	free(src[0]);
	free(src[1]);

	return 0;
}
